#!/usr/bin/python3
# -*- coding: utf-8 -*

from dwarf_values import *
import struct
import sys

try:
    with open(sys.argv[1], 'rb') as f:
        OPCODES = list(f.read())
except:
    sys.exit('usage: %s /path/to/program.bin' % sys.argv[0])

OFFSET = 35

op_ptr = OFFSET

def read_2s(op_ptr):
    s = bytearray(OPCODES[op_ptr:op_ptr+2])
    return struct.unpack('<h', s)[0]

def read_1u(op_ptr):
    return OPCODES[op_ptr]

def read_2u(op_ptr):
    return OPCODES[op_ptr]|OPCODES[op_ptr+1] << 8

def read_4u(op_ptr):
    return (OPCODES[op_ptr]
        | OPCODES[op_ptr+1] << 8
        | OPCODES[op_ptr+2] << 16
        | OPCODES[op_ptr+3] << 24)

def read_8u(op_ptr):
    return (OPCODES[op_ptr]
        | OPCODES[op_ptr+1] << 8
        | OPCODES[op_ptr+2] << 16
        | OPCODES[op_ptr+3] << 24
        | OPCODES[op_ptr+4] << 32
        | OPCODES[op_ptr+5] << 40
        | OPCODES[op_ptr+6] << 48
        | OPCODES[op_ptr+7] << 56)

while op_ptr < 0x172b:
    op_ptr_orig = op_ptr
    s = ''
    opcode = OPCODES[op_ptr]
    op_ptr += 1
    if opcode == DW_OP_bra: # 0x28
        offset = read_2s (op_ptr)
        op_ptr += 2
        op_ptr_next = op_ptr + offset
        s = 'BRANCH %s (offset=%d)\n' % (hex(op_ptr_next), offset)
    elif opcode == DW_OP_pick:
        offset = OPCODES[op_ptr]
        op_ptr += 1
        s = 'PICK offset = %d' % (offset)
    elif opcode == DW_OP_skip:
        offset = read_2s (op_ptr)
        op_ptr += 2
        op_ptr_next = op_ptr + offset
        s = 'GOTO %s (offset=%d)\n' % (hex(op_ptr_next), offset)
    elif (opcode == DW_OP_lit0
        or opcode == DW_OP_lit1
        or opcode == DW_OP_lit2
        or opcode == DW_OP_lit3
        or opcode == DW_OP_lit4
        or opcode == DW_OP_lit5
        or opcode == DW_OP_lit6
        or opcode == DW_OP_lit7
        or opcode == DW_OP_lit8
        or opcode == DW_OP_lit9
        or opcode == DW_OP_lit10
        or opcode == DW_OP_lit11
        or opcode == DW_OP_lit12
        or opcode == DW_OP_lit13
        or opcode == DW_OP_lit14
        or opcode == DW_OP_lit15
        or opcode == DW_OP_lit16
        or opcode == DW_OP_lit17
        or opcode == DW_OP_lit18
        or opcode == DW_OP_lit19
        or opcode == DW_OP_lit20
        or opcode == DW_OP_lit21
        or opcode == DW_OP_lit22
        or opcode == DW_OP_lit23
        or opcode == DW_OP_lit24
        or opcode == DW_OP_lit25
        or opcode == DW_OP_lit26
        or opcode == DW_OP_lit27
        or opcode == DW_OP_lit28
        or opcode == DW_OP_lit29
        or opcode == DW_OP_lit30
        or opcode == DW_OP_lit31):
        result = opcode - DW_OP_lit0
        s = 'LIT %d' % (result)
    elif opcode == DW_OP_rot:
        s = 'ROT'
    elif opcode == DW_OP_dup:
        s = 'DUP'
    elif opcode == DW_OP_const8u:
        result = read_8u(op_ptr)
        op_ptr += 8
        s = 'CONST8U %s' % (hex(result))
    elif opcode == DW_OP_const4u:
        result = read_4u(op_ptr)
        s = 'CONST4U %s' % (hex(result))
        op_ptr += 4
    elif opcode == DW_OP_and:
        s = 'AND'
    elif opcode == DW_OP_swap:
        s = 'SWAP'
    elif opcode == DW_OP_const1u:
        result = read_1u(op_ptr)
        op_ptr += 1
        s = 'CONST1U %s' % (hex(result))
    elif opcode == DW_OP_shr:
        s = 'SHR'
    elif opcode == DW_OP_plus:
        s = 'PLUS'
    elif opcode == DW_OP_xor:
        s = 'XOR'
    elif opcode == DW_OP_minus:
        s = 'MINUS'
    elif opcode == DW_OP_shl:
        s = 'SHL'
    elif opcode == DW_OP_deref_size:
        opcode2 = OPCODES[op_ptr]
        op_ptr += 1
        s = 'DEREF SIZE %d' % (opcode2)
    elif opcode == DW_OP_or:
        s = 'OR'
    elif opcode == DW_OP_mul:
        s = 'MUL'
    elif opcode == DW_OP_drop:
        s = 'DROP'
    elif opcode == DW_OP_deref:
        s = 'DEREF'
    elif opcode == DW_OP_nop:
        s = 'NOP'
    elif opcode == 0:
        s = 'NIL'
    elif opcode > DW_OP_nop:
        s = 'UNKNOWN %s ' % (hex(opcode))
    else:
        s = 'UNKNOWN %s ' % (hex(opcode))
    print('%s: %s' % (hex(op_ptr_orig), s))
