#!/usr/bin/python3
# -*- coding: utf-8 -*

import sys

try:
    with open(sys.argv[1], 'rb') as f:
        OPCODES = list(f.read())
except:
    sys.exit('usage: %s /path/to/program.bin' % sys.argv[0])

def read_4u(address):
    if address >= 0x400258:
        address -= 0x400258
    else:
        raise Exception
    return (OPCODES[address]
        | OPCODES[address+1] << 8
        | OPCODES[address+2] << 16
        | OPCODES[address+3] << 24)

def read_8u(address):
    if address >= 0x400258:
        address -= 0x400258
    else:
        raise Exception
    return (OPCODES[address]
        | OPCODES[address+1] << 8
        | OPCODES[address+2] << 16
        | OPCODES[address+3] << 24
        | OPCODES[address+4] << 32
        | OPCODES[address+5] << 40
        | OPCODES[address+6] << 48
        | OPCODES[address+7] << 56)

def moulinette2(A, B):
    for i in range(0, 4):
        B_hi = (((B & 0xffffffff00000000)) >> 0x20)
        A_hi = (((A & 0xffffffff00000000)) >> 0x20)
        A_lo = (A & 0xffffffff)
        B_lo = (B & 0xffffffff)
        value = read_4u((((B_hi & 0xff)) << 0x2) + 0x4006b4)
        tmp = ((A_lo ^ (A_hi + B_lo)) & 0xffffffff)
        A2_lo = ((tmp + value) & 0xffffffff)
        A2_hi = ((A_hi & B_lo))
        B2_lo = ((B_lo - tmp) & 0xffffffff)
        B2_hi = (A2_lo ^ ((B_hi >> 0x8)))
        A2 = (A2_lo|(A2_hi << 0x20))
        B2 = (B2_lo|(B2_hi << 0x20))
        A = A2
        B = B2
    return (A, B)

def fun_0x23(A, B, C, D):
    X0 = A
    X1 = B
    X2 = C
    X3 = D
    R2 = 0
    R3 = 0
    R4 = 0
    R5 = 0
    R6 = 0
    R7 = 0
    cnt4 = 0
    R0 = C
    R1 = D
    which = 0
    S1 = A
    S0 = B
    cnt15 = 0
    while cnt4 != 4:
        if which == 1:
            X0 = R0
            X1 = R1
        else:
            X2 = R0
            X3 = R1
        (R0, R1) = moulinette2(R0, R1)
        if which == 0:
            R2 = X0
            R3 = X1
        else:
            R2 = X2
            R3 = X3
        cnt15 = 0
        (R4, R5) = fun_0x154(0, R0, R1)
        R6 = R3 & 0xffffffff
        R7 = (R3 & 0xffffffff00000000) >> 0x20
        e = 0
        while cnt15 != 15:
            (R6, R7) = moulinette(R6, R7)
            if e != 0:
                assert(e == 1)
                R3_hi = (((R3 & 0xffffffff00000000)) >> 0x20)
                R3_lo = (R3 & 0xffffffff)
                F = ((((((R3_hi ^ R6) ^ R7)) >> 0xe) | ((((R3_hi ^ R6) ^ R7)) << 0x12)) & 0xffffffff)
                G = (((((R3_lo ^ R7)) << 0x1a) | (((R3_lo ^ R7)) >> 0x6)) & 0xffffffff)
                e = 0
                R3 = (R5 ^ (G | ((F) << 0x20)))
                R6 = (R3 & 0xffffffff)
                R7 = (((R3 & 0xffffffff00000000)) >> 0x20)
                cnt15 += 1
                (R4, R5) = fun_0x154(cnt15, R0, R1)
            else:
                R2_hi = (((R2 & 0xffffffff00000000)) >> 0x20)
                R2_lo = (R2 & 0xffffffff)
                E = ((((((R2_hi ^ R7)) >> 0x12) | (((R2_hi ^ R7)) << 0xe)) & 0xffffffff) ^ R6)
                D = ((((((R2_lo ^ R6)) << 0x4) | (((R2_lo ^ R6)) >> 0x1c)) & 0xffffffff) ^ R7)
                e = 0x1
                R2 = (R4 ^ (D | ((E) << 0x20)))
                R6 = (R2 & 0xffffffff)
                R7 = (((R2 & 0xffffffff00000000)) >> 0x20)
            # second while end
        # first while indentation
        if which != 0:
            cnt4 += 1
            R0 = R2
            R1 = R3
            which = 0
        else:
            which = 1
            S1 = R2
            S0 = R3
            R0 = R2
            R1 = R3
        # first while end
    return (S1, S0, R0, R1)

def fun_0x154(cnt, A, B):
    for C in range(0, cnt+1):
        value = read_4u((C * 0x4) + 0x400678)
        I = (((((A & 0xffffffff00000000)) >> 0x20) + 0x45786532) & 0xffffffff)
        G = (I ^ (B & 0xffffffff))
        H = (((((((B & 0xffffffff00000000)) >> 0x20)) << 0x4) | (((((B & 0xffffffff00000000)) >> 0x20)) >> 0x1c)) & 0xffffffff)
        J = (((value ^ (A & 0xffffffff)) - I) & 0xffffffff)
        e = (G & 0x80000000)
        K = 0
        if e == 0:
            K = (0x818f694a)
        else:
            K = (0x60bf080f)
        A = (J | ((((K ^ I) ^ H)) << 0x20))
        B = (G | ((H) << 0x20))
    return (A, B)

def moulinette(R6, R7):
    for idx in range(0, 6):
        value1 = read_4u(((idx * 0x2) * 0x4) + 0x400648)
        value2 = read_4u((((idx * 0x2) + 0x1) * 0x4) + 0x400648)
        tmp = ((R7 + value1) & 0xffffffff)
        R6 = R6 ^ tmp
        R7 = (((value2 | R6) ^ R7) & 0xffffffff)
    return (R6, R7)

def inv1(R0, R1, R2, R3):
    for cnt in reversed(range(0, 15)):
        (R4, R5) = fun_0x154(cnt, R0, R1)
        # R3 en fin de 2n + 1 donne F et G
        #F = ((R3 ^ R5) >> 0x20) & 0xffffffff
        F = ((((R3 ^ R5) & 0xffffffff00000000)) >> 0x20)
        G = (R3 ^ R5) & 0xffffffff
        # R2 en fin de 2n donne D et E
        #E = ((R2 ^ R4) >> 0x20) & 0xffffffff
        E = ((((R2 ^ R4) & 0xffffffff00000000)) >> 0x20)
        D = (R2 ^ R4) & 0xffffffff
        # R6 et R7 en fin de 2n
        R6 = (R2 & 0xffffffff)
        R7 = (((R2 & 0xffffffff00000000)) >> 0x20)
        # R6 et R7 en debut de 2n+1
        (R6, R7) = moulinette(R6, R7)
        # R3 en fin de 2n-1
        R3_lo = (((G >> 0x1a) | (G << 0x6)) & 0xffffffff) ^ R7
        R3_hi = ((((F << 0xe) | (F >> 0x12)) & 0xffffffff) ^ R6 ^ R7) & 0xffffffff
        R3 = R3_lo | (R3_hi << 0x20)
        # R6 et R7 en fin de 2n-1
        R6 = (R3 & 0xffffffff)
        R7 = (((R3 & 0xffffffff00000000)) >> 0x20)
        # R6 et R7 en debut de 2n
        (R6, R7) = moulinette(R6, R7)
        R2_hi = ((((E ^ R6) << 0x12) | ((E ^ R6) >> 0xe)) ^ R7) & 0xffffffff
        R2_lo = ((((D ^ R7) >> 0x4) | ((D ^ R7) << 0x1c)) ^ R6) & 0xffffffff
        R2 = R2_lo | (R2_hi << 0x20)
    return (R2, R3)

def int_to_bytes(x):
    return x.to_bytes((x.bit_length() + 7) // 8, 'little')

def inv2(R0, R1, R2, R3):
    for cnt4 in reversed(range(0, 4)):

        # input 2n+1:
        I0 = R0
        I1 = R1
        (I0, I1) = moulinette2(I0, I1)
        (I2, I3) = inv1(I0, I1, R2, R3)

        # I2, I3 est input de 2n = output de 2n-1
        (I4, I5) = moulinette2(I2, I3)
        (I0, I1) = inv1(I4, I5, R0, R1)
        R0 = I0
        R1 = I1
        R2 = I2
        R3 = I3
    print(int_to_bytes(R0)+int_to_bytes(R1)+int_to_bytes(R2)+int_to_bytes(R3))

def int_from_bytes(xbytes):
    return int.from_bytes(bytes(xbytes, 'utf8'), 'little')

# Valeurs de test
#test = fun_0x23(int_from_bytes('00112233'), int_from_bytes('44556677'),
#    int_from_bytes('8899AABB'), int_from_bytes('CCDDEEFF'))
#inv2(test[0], test[1], test[2], test[3])

inv2(0x65850b36e76aaed5, 0xd9c69b74a86ec613, 0xdc7564f1612e5347, 0x658302a68e8e1c24)
