#!/usr/bin/python3
# -*- coding: utf-8 -*

import sys

try:
    with open(sys.argv[1], 'rb') as f:
        program = list(f.read())
except:
    sys.exit('usage: %s /path/to/program.bin' % sys.argv[0])

def print_opcode(pc, opcode, s):
    # reorder opcode
    opcode_str = '%02x%02x%02x' % (opcode & 0xff, (opcode >> 8) & 0xff, (opcode
        >> 16) & 0xff)
    print('%06x: %s %s' % (pc, opcode_str, s))

def disassemble(program):
    pc = 0
    while pc < 0x1b9:
        s = None
        opcode = (program[pc]
            | (program[pc + 1] << 8)
            | (program[pc + 2] << 16))
        c = (opcode >> 0x14) & 0xf
        if (c == 0xc):
            dst = (opcode >> 0xe) & 0xf
            s = 'cle = f(r%d)' % dst

        elif (c == 0xa):
            s = 'ret'
        elif (c == 0xd):
            s = '*(0x9010000+8) = 0'
        elif (c == 0xe):
            imm = opcode & 0x3fff
            s = 'jmp 0x%x si (*(0x9010000) & 0xffffffff) > 5' % imm
        else:
            c2 = (opcode >> 0x12) & 0x3
            dst = (opcode >> 0xe) & 0xf
            if c2 == 0:
                # 0xe2007c0
                src = (opcode >> 0xa) & 0xf
                if (c == 0x0):
                    s = 'mov r%d, r%d' % (dst, src)
                elif (c == 0x1):
                    s = 'dec r%d' % dst # 'smc(0x%x, r%d, r%d) puis add' % (0x83010011, dst, src)
                elif (c == 0x2):
                    s = 'add r%d, r%d' % (dst, src) # 'smc(0x%x, r%d, r%d)' % (0x83010012, dst, src)
                elif (c == 0x3):
                    s = 'sub r%d, r%d' % (dst, src) # 'smc(0x%x, r%d, r%d)' % (0x83010013, dst, src)
                elif (c == 0x6):
                    s = 'xor r%d, r%d' % (dst, src) # 'smc(0x%x, r%d, r%d)' % (0x83010016, dst, src)
                elif (c == 0xb):
                    s = 'inv(r%d)' % dst # 'et and'
            elif c2 == 0x3:
                # 0xe200938
                imm = opcode & 0x3fff
                if (c == 0x0):
                    s = 'mov r%d, 0x%x' % (dst, imm)
                elif (c == 0x2):
                    s = 'add r%d, 0x%x' % (dst, imm) # 'smc(0x%x, r%d, 0x%x)' % (0x83010022, dst, imm)
                elif (c == 0x3):
                    s = 'sub r%d, 0x%x' % (dst, imm) # 'smc(0x%x, r%d, 0x%x)' % (0x83010023, dst, imm)
                elif (c == 0x4):
                    s = 'lsl r%d, 0x%x' % (dst, imm)
                elif (c == 0x5):
                    s = 'lsr r%d, 0x%x' % (dst, imm)
                elif (c == 0x6):
                    s = 'smc(0x%x, r%d, 0x%x)' % (0x83010026, dst, imm)
                elif (c == 0x7):
                    s = 'and r%d, 0x%x' % (dst, imm) # 'smc(0x%x, r%d, 0x%x)' % (0x83010027, dst, imm)
                elif (c == 0x8):
                    s = 'goto 0x%x' % imm # 'smc(0x%x, r%d, 0x%x)' % (0x83010028, dst, imm)
                elif (c == 0x9):
                    s = 'jmp 0x%x si r%d est non nul' % (imm, dst)
            elif c2 == 0x1:
                # 0xe200b6c
                assert(c == 0x0)
                src = (opcode >> 0xa) & 0xf
                s = 'ldr r%d, [r%d]' % (dst, src)
            elif c2 == 0x2:
                # 0xe200bb8
                assert(c == 0x0)
                src = (opcode >> 0xa) & 0xf
                s = 'str r%d, [r%d]' % (src, dst)

        if s == None:
            s = 'unk'
        print_opcode(pc, opcode, s)
        pc += 3

disassemble(program)
