#!/usr/bin/python3
# -*- coding: utf-8 -*

import sys

try:
    with open(sys.argv[1], 'rb') as f:
        program = list(f.read())
except:
    sys.exit('usage: %s /path/to/program.bin' % sys.argv[0])

def read_1u(pc, prog):
    v = prog[pc]
    return v & 0xff

def read_2u(pc, prog):
    v = (prog[pc]
        | (prog[pc + 1] << 8))
    return v & 0xffff

def read_4u(pc, prog):
    v = (prog[pc]
        | (prog[pc + 1] << 8)
        | (prog[pc + 2] << 16)
        | (prog[pc + 3] << 24))
    return v & 0xffffffff

def set_value(pc, v, prog):
    prog[pc] = v & 0xff
    prog[pc + 1] = (v & 0xff00) >> 8 & 0xff
    prog[pc + 2] = (v & 0xff0000) >> 16 & 0xff
    prog[pc + 3] = (v & 0xff000000) >> 24 & 0xff

def set_value2(pc, v, prog):
    set_value(pc, v, prog)

def inv(r0):
    return (((r0 & 0xff) << 0x8) | ((r0 & 0xff00) >> 0x8)) & 0xffff

def set_key(r, prog):
    v = (r >> 2) & 0x7
    if v == 1:
        set_value2(0x100004, 0x6766722e, prog)
    elif v == 2:
        set_value2(0x100008, 0x666e632e, prog)
    elif v == 3:
        set_value2(0x10000c, 0x2e76662e, prog)
    elif v == 4:
        set_value2(0x100010, 0x76706e73, prog)
    elif v == 5:
        set_value2(0x100014, 0x66407279, prog)
    elif v == 6:
        set_value2(0x100018, 0x70766766, prog)
    elif v == 7:
        set_value2(0x10001c, 0x7465622e, prog)
    else:
        set_value2(0x100000, 0x612e7270, prog)

def first_run(prog, r7_list, prog_list):
    r4 = 0x10
    r4 = (r4 << 0x10) & 0xffffffff
    r4 += 0x20
    r13 = 0x10
    r13 = (r13 << 0x10) & 0xffffffff
    r13 += 0x20
    for r12 in range(0, 4):
        r0 = inv(read_2u(r4, prog))
        r4 += 0x2
        r1 = inv(read_2u(r4, prog))
        r4 += 0x2
        r2 = inv(read_2u(r4, prog))
        r4 += 0x2
        r3 = inv(read_2u(r4, prog))
        r14 = 0x20
        r7 = 0x7
        while r14 != 0:
            r14 -= 1
            r7_list[r12][r14] = r7
            prog_list[r12][r14] = list(prog)
            (r7, r9) = f(r1, r7, prog)
            r8 = (r14 >> 0x3) & 0x1
            set_key(r14, prog)
            if r8 == 0:
                r3_prev = r3
                r3 = (r14 + 0x1) ^ r0 ^ r1
                r0 = r9
                r1 = r2
                r2 = r3_prev
            else:
                r0_prev = r0
                r0 = r9
                r1 = (r14 + 0x1) ^ r0 ^ r2
                r2 = r3
                r3 = r0_prev
        r0 = inv(r0)
        r1 = inv(r1)
        r1 = (r1 << 0x10) & 0xffffffff
        r0 += r1
        set_value(r13, r0, prog)
        r13 += 0x4
        r2 = inv(r2)
        r2 = inv(r3)
        r3 = (r3 << 0x10) & 0xffffffff
        r2 += r3
        set_value(r13, r2, prog)
        r13 += 0x4
        r4 = r13

def f(r1, r7, prog):
    r4 = (r1 >> 0x8) & 0xff
    r10 = (r7 << 0x10) + ((r1 & 0xff) << 0x8) + ((r1 >> 0x8) & 0xff) + 0x1000
    r6 = read_1u(r10, prog) & 0xff
    if r7 == 0:
        r7 = 0xa
    r7 -= 1
    r10 = (r7 << 0x10) + (r1 & 0xff00) + r6 + 0x1000
    r5 = read_1u(r10, prog) & 0xff
    if r7 == 0:
        r7 = 0xa
    r7 -= 1
    r10 = (r7 << 0x10) + (r6 << 0x8) + r5 + 0x1000
    r4 = read_1u(r10, prog) & 0xff
    if r7 == 0:
        r7 = 0xa
    r7 -= 1
    r10 = (r7 << 0x10) + (r5 << 0x8) + r4 + 0x1000
    r6 = read_1u(r10, prog) & 0xff
    if r7 == 0:
        r7 = 0xa
    r7 -= 1
    r9 = (r6 << 0x8) + r4
    return (r7, r9)

def reverse(i, r0, r1, r2, r3, r14):
    if r14 == 32:
        r0 = inv(r0)
        r1 = inv(r1)
        r2 = inv(r2)
        r3 = inv(r3)
        return (r0, r1, r2, r3)
    possible = []
    for r1_prev in range(0, 0x10000):
        #print 'trying', i, r14, hex(r1)
        r7 = r7_list[i][r14]
        prog = prog_list[i][r14]
        if f(r1_prev, r7, prog)[1] == r0:
            possible.append(r1_prev)
    assert(len(possible) == 1)
    #print 'possible', [hex(e) for e in possible]
    for r1_prev in possible:
        if ((r14 >> 0x3) & 1) == 0:
            r0_prev = r1_prev ^ r3 ^ (r14 + 0x1)
            r2_prev = r1
            r3_prev = r2
            return reverse(i, r0_prev, r1_prev, r2_prev, r3_prev, r14+0x1)
        else:
            r0_prev = r3
            r2_prev = r1 ^ r0 ^ (r14 + 0x1)
            r3_prev = r2
            return reverse(i, r0_prev, r1_prev, r2_prev, r3_prev, r14+0x1)

def int_to_bytes(x):
    return x.to_bytes((x.bit_length() + 7) // 8, 'little')

r7_list = 4 * [32 * [0]]
prog_list = 4 * [32 * [None]]
p2 = list(program)
first_run(p2, r7_list, prog_list)

result = []
for i in range(0, 4):
    expected1 = read_4u(0x100000+i*8, p2)
    expected2 = read_4u(0x100000+i*8+4, p2)

    r0 = inv(expected1 & 0xffff) & 0xffff
    r1 = inv((expected1 & 0xffff0000) >> 0x10) & 0xffff
    r2 = inv(expected2 & 0xffff) & 0xffff
    r3 = inv((expected2 & 0xffff0000) >> 0x10) & 0xffff

    (r0, r1, r2, r3) = reverse(i, r0, r1, r2, r3, 0)

    input1 = r0 | (r1 << 0x10) & 0xffffffff
    input2 = r2 | (r3 << 0x10) & 0xffffffff

    result.extend([input1, input2])

s = ''
for v in result:
    s = s + (int_to_bytes(v)).hex()
print(s)
